/*
 * RiskScape™ Copyright New Zealand Institute for Earth Science Limited
 * (Earth Sciences New Zealand) is distributed for research purposes only
 * under the terms of AGPLv3.
 *
 * RiskScape™ Copyright 2026 New Zealand Institute for Earth Science
 * Limited (Earth Sciences New Zealand). All rights reserved. Source code
 * available under the AGPLv3.
 * 
 * This program is free software: you can redistribute it and/or modify it under
 *  the terms of the GNU Affero General Public License as published by the Free
 *  Software Foundation, either version 3 of the License, or (at your option) any
 *  later version.
 * 
 * This program is distributed for RESEARCH PURPOSES ONLY, in the hope that it will
 * be useful for research and education initiatives.
 * 
 * If you are not a researcher, or you are a researcher who wishes to use this
 * program on terms other than AGPLv3 (including those who wish to restrict the
 * distribution of any source code created using this program), please contact:
 * https://riskscape.org.nz
 * 
 * This program is distributed WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Affero General Public License for more details.  You should have received a copy
 * of the GNU Affero General Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 * 
 * By way of summary only, under the AGPLv3:
 *     • Permissions of this strongest copyleft license are conditioned
 *       on making available complete source code of licensed works and
 *       modifications, which include larger works using a licensed work,
 *       under the same license.
 *     • Copyright and license notices must be preserved.
 *     • Contributors provide an express grant of patent rights.
 *     • When a modified version is used to provide a service over a
 *       network, the complete source code of the modified version must be made
 *       available.
 */
package nz.org.riskscape.engine.function;

import static org.junit.Assert.*;

import java.util.Collections;

import org.junit.Test;

import com.google.common.collect.Lists;

public class MathsFunctionsTest {

  @Test
  public void canCreateAPolynomialFunction() {
    RiskscapeFunction function = Maths.newPolynomial(new double[] {2, 1});

    assertEquals(3D, function.call(Collections.singletonList(1D)));
  }

  @Test
  public void canCreateAPolynomialFunctionWithDynamicParameters() {
    RiskscapeFunction function = Maths.newPolynomial(2);

    assertEquals(3D, function.call(Lists.newArrayList(1D, 2, 1)));
  }

  @Test
  public void canCreateAPowerFunction() {
    RiskscapeFunction function = Maths.newPower(2);

    assertEquals(4D, function.call(Collections.singletonList(2D)));
  }

  @Test
  public void canCreateAPowerFunctionWithDynamicParameters() {
    RiskscapeFunction function = Maths.newPower();

    assertEquals(4D, function.call(Lists.newArrayList(2D, 2)));
    assertEquals(8D, function.call(Lists.newArrayList(2D, 3)));
  }

  @Test
  public void canCreateANormalDistrib() {
    RiskscapeFunction function = Maths.newCumulativeNormalDistribution(2, 1);

    assertEquals(0.5D, function.call(Collections.singletonList(2D)));
  }

  @Test
  public void canCreateANormalDistribWithDynamicParameters() {
    RiskscapeFunction function = Maths.newCumulativeNormalDistribution();

    assertEquals(0.5D, function.call(Lists.newArrayList(2D, 2, 1)));
  }

  @Test
  public void canCreateALogNormalDistrib() {
    RiskscapeFunction function = Maths.newCumulativeLogNormalDistribution(0, 10);

    assertEquals(0.5D, function.call(Collections.singletonList(1D)));
  }

  @Test
  public void canCreateALogNormalDistribWithDynamicParameters() {
    RiskscapeFunction function = Maths.newCumulativeLogNormalDistribution();

    assertEquals(0.5D, function.call(Lists.newArrayList(1D, 0, 10)));
  }
}
